/*  WB-tree File Based Associative String Data Base System. */
/*  Copyright (c) 1991, 1992, 1993, 2000 Free Software Foundation, Inc. */
/*  */
/*  This program is free software; you can redistribute it and/or modify */
/*  it under the terms of the GNU General Public License as published by */
/*  the Free Software Foundation; either version 2, or (at your option) */
/*  any later version. */
/*  */
/*  This program is distributed in the hope that it will be useful, */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/*  GNU General Public License for more details. */
/*  */
/*  You should have received a copy of the GNU General Public License */
/*  along with this software; see the file COPYING.  If not, write to */
/*  the Free Software Foundation, 59 Temple Place, Suite 330, Boston, */
/*  MA 02111, USA. */
/*  */
/*  As a special exception, the Free Software Foundation gives permission */
/*  for additional uses of the text contained in its release of GUILE. */
/*  */
/*  The exception is that, if you link the GUILE library with other files */
/*  to produce an executable, this does not by itself cause the */
/*  resulting executable to be covered by the GNU General Public License. */
/*  Your use of that executable is in no way restricted on account of */
/*  linking the GUILE library code into it. */
/*  */
/*  This exception does not however invalidate any other reasons why */
/*  the executable file might be covered by the GNU General Public License. */
/*  */
/*  This exception applies only to the code released by the */
/*  Free Software Foundation under the name GUILE.  If you copy */
/*  code from other Free Software Foundation releases into a copy of */
/*  GUILE, as the General Public License permits, the exception does */
/*  not apply to the code that you add in this way.  To avoid misleading */
/*  anyone as to the status of such modified files, you must delete */
/*  this exception notice from them. */
/*  */
/*  If you write modifications of your own for GUILE, it is your choice */
/*  whether to permit this exception to apply to your modifications. */
/*  If you do not wish that, delete this exception notice. */


#include "wbsys.h"
/* @code{bt_scan} scans all keys in the range [@var{key1}..@var{key2}), */
/* performing one of several functions: */
/*  */
/* @multitable @columnfractions .2 .15 .66 */
/* @item @var{operation} */
/* @tab @var{func} */
/* @tab RESULT */
/* @item COUNT-SCAN */
/* @tab NIL */
/* @tab counts all keys in range */
/* @item COUNT-SCAN */
/* @tab given */
/* @tab counts all keys in range satisfying @var{func} */
/* @item REM-SCAN */
/* @tab NIL */
/* @tab deletes all keys in range */
/* @item REM-SCAN */
/* @tab given */
/* @tab deletes all keys in range satisfying @var{func} */
/* @item MODIFY-SCAN */
/* @tab NIL */
/* @tab ARGERR */
/* @item MODIFY-SCAN */
/* @tab given */
/* @tab updates values for keys in range satisfying @var{func} */
/* @end multitable */
/*  */
/* @code{bt_scan} returns SUCCESS if scan completed; under any other result code */
/* the scan is resumable. The possible results are: */
/*  */
/* @table @asis */
/* @item NOTPRES */
/* meaning the @var{blk_limit} was exceeded; */
/* @item RETRYERR */
/* meaning @var{func} or delete got a RETRYERRR; */
/* @item TERMINATED */
/* meaning @var{func} asked to terminate the scan; */
/* @item <other error> */
/* means @var{func} or DELETE encountered this errror. */
/* @end table */
/*  */
/* Each block of data is scanned/deleted/modified in a single operation */
/* that is, the block is found and locked only once, and only written after */
/* all modifications are made. Tho only exception is that MODIFY-SCANs */
/* that increase the size of values  can cause block splits. Such cases */
/* are detected and converted to a PUT plus a NEXT. This has */
/* two consequences: data is written out each time a PUT occurs, */
/* and it is conceivable that @var{func} may be called more than once on the */
/* key value that caused the split if a RETRYERR occurs in the PUT. */
/* However, SCAN guarantees that only one modification will actually be */
/* made in this case (so that one can write INCREMENT-RANGE, for example). */
/*  */
/* @var{func} is passed pointers to (copies of) the key and value, */
/* plus one user argument: */
/* @example */
/* @var{func} (keystr klen vstr vlen extra_arg); */
/* @end example */
/* @var{func} is expected to return either: SUCCESS for DELETE/COUNT, */
/* NOTPRES/NOTDONE for SKIP (ie, DONT DELETE/COUNT), or */
/* any other code to terminate the scan resumably at the current point. */
/* For MODIFY-SCAN, if changing the value, the new value length is returned. */
/* Except for the case mentioned above, the caller can depend on @var{func} */
/* being called exactly once for each key value in the specified range, */
/* and only on those values. */
/*  */
/* If @var{key2} <= @var{key1}, then no scan will occur (even if @var{key1} is found). */
/* To make possible bounded-time operation @code{bt_scan} will */
/* access at most @var{blk_limit} blocks at a time; if you dont care, */
/* give it -1 for @var{blk_limit}. */
/*  */
/* The number of keys deleted/counted/modified is returned in the */
/* @code{skey-count} field of @var{respkt}; the key to resume at is returned in */
/* @var{key1} (@emph{which therefore needs to be 256 bytes long}); and the */
/* new key length is returned in the @code{skey-len} field of @var{respkt}. If */
/* returns SUCCESS, @code{skey-len} is zero.  NOTE that */
/* @code{skey-count} is cumulative, so the caller needs to initialize */
/* it to 0 when starting a new @code{bt_scan}. */
/*  */
/* @emph{WARNING:} when @code{bt_scan} returns other than SUCCESS, it modifies */
/* the @var{key1} string so that the string args are correctly set up for the */
/* next call (the returned value is the new length for @var{key1}). */
/* Therefore, @emph{@var{key1} must be a maximum-length string!} */

int bt_scan(han, operation, key1, len1, key2, len2, func, long_tab, respkt, blk_limit)
     HAND *han;
     int operation;
     unsigned char *key1;
     int len1;
     unsigned char *key2;
     int len2;
     int_function func;
     long *long_tab;
     int *respkt;
     int blk_limit;
{
L_bt_scan:
  {
    int pkt[pkt_size];
    int opkt[pkt_size];
    ENTRY *ent = 0;
    unsigned char vstr[0x100];
    int accmode = (operation)==(count_scan)
      ?accread
      :accwrite;
    int result = success;
    if ((len1)<-2)
      {
	dprintf((diagout, ">>>>ERROR<<<< %s: bad length string1 %d\n", "bt_scan", len1));
	return argerr;
      }
    else if ((len2)<-1)
      {
	dprintf((diagout, ">>>>ERROR<<<< %s: bad length string2 %d\n", "bt_scan", len2));
	return argerr;
      }
    else if (((operation)==(modify_scan))
        && (!(func)))
      {
	dprintf((diagout,
	 ">>>>ERROR<<<< %s: MODIFY-SCAN requires func be specified\n",
	 "bt_scan"));
	return argerr;
      }
    else {
      ent = chain_find_ent(han, accmode, key1, len1, pkt);
      if ((ent)
	  && (blk_find_pos(ent_blk(ent), key2, len2, opkt)))
	{
	  if ((operation)==(count_scan))
	    {
	      ENTRY *nent = allocate_ent();
	      ent_copy(nent, ent);
	      release_ent(ent, accmode);
	      result = chain_scan(nent, operation, pkt, opkt, key1, func, long_tab, vstr, respkt, han_wcb(han));
	      recycle_ent(nent);
	    }
	  else {
	    result = chain_scan(ent, operation, pkt, opkt, key1, func, long_tab, vstr, respkt, han_wcb(han));
	    release_ent(ent, accmode);
	    if ((result)>0)
	      {
		result = bt_put(han, key1, skey_len(respkt), vstr, result);
		if ((result)==(success))
		  {
		    set_skey_count(respkt, (skey_count(respkt))+1);
		    set_skey_len(respkt, increment_string(key1, skey_len(respkt), 0x100));
		    result = notpres;
		  }
	      }
	  }
	  if (((result)==(notpres))
	      && (0!=(blk_limit)))
	    {
	      len1 = skey_len(respkt);
	      blk_limit = (blk_limit)-1;
	      goto L_bt_scan;
	    }
	  else return result;
	}
      else {
	if (ent)
	  release_ent(ent, accmode);
	rem_fct = 0x1L+(rem_fct);
	return unkerr;
      }
    }
  }
}

/*  this function increments a string lexicographically */

int increment_string(str, len, maxlen)
     unsigned char *str;
     int len;
     int maxlen;
{
  if ((len)<(maxlen))
    {
      str[len] = (0);
      return (len)+1;
    }
  else {
    int oldval = (unsigned)(str[(len)-1]);
    str[(len)-1] = (1+(oldval));
    return len;
  }
}

/*  Each call to CHAIN-SCAN scans */
/*  all the keys within the specified range WITHIN block ENT. */
/*  If the scan actually reaches  the end of range, it sets SKEY-LEN=0 */
/*  and returns SUCCESS. If there's more to the range, */
/*  it sets KEY-STR to the key to continue deleting */
/*  from (ie, the split key of ENT), SKEY-LEN to its length, and */
/*  returns NOTPRES (NOTDONE). The caller must then call CHAIN-FIND */
/*  to find the START and END keys and call again. */


int chain_scan(ent, operation, pkt, opkt, key_str, func, long_tab, vstr, respkt, wcb)
     ENTRY *ent;
     int operation;
     int *pkt;
     int *opkt;
     unsigned char *key_str;
     int_function func;
     long *long_tab;
     unsigned char *vstr;
     int *respkt;
     int wcb;
{
  {
    unsigned char *blk = ent_blk(ent);
    int result = success;
    if (((operation)==(rem_scan))
        && (!(func))
        && ((match_pos(opkt))>(match_pos(pkt)))
        && ((match_pos(pkt))==(blk_data_start))
        && (at_split_key_pos_P(blk, match_pos(opkt))))
      {
	{
	  int key_len = recon_this_key(blk, match_pos(opkt), key_str, 0, 0x100);
	  substring_move(key_str, 0, key_len, blk, (blk_data_start)+2);
	  set_field_len(blk, (blk_data_start)+1, key_len);
	  blk_set_end(blk, (blk_data_start)+2+(key_len));
	}
	set_skey_count(respkt, (skey_count(respkt))+1);
	rem_ct = 0x1L+(rem_ct);
	ent_set_dty(ent, !0);
	set_match_pos(opkt, blk_data_start);
      }
    else {
      int oldct = skey_count(respkt);
      unsigned char ckstr[0x100];
      int clen = 0;
      if (func)
	clen = recon_this_key(blk, match_pos(pkt), ckstr, 0, 0x100);
      set_match_type(pkt, match);
      result = scan_loop(ent_blk(ent), operation, pkt, opkt, func, long_tab, respkt, ckstr, clen, vstr, seg_bsiz(ent_seg(ent)));
      if (((operation)!=(count_scan))
	  && ((skey_count(respkt))>(oldct)))
	ent_set_dty(ent, !0);
    }
    if (((operation)==(rem_scan))
        && (blk_empty_P(blk))
        && (!(end_of_chain_P(blk))))
      {
	blk_delete(ent);
      }
    else if (ent_dty_P(ent))
      if ((((operation)==(rem_scan))
	   && ((wcb_sar_P(wcb))
	       || ((blk_level(blk))>(leaf))))
	  || (((operation)==(modify_scan))
	      && (wcb_sap_P(wcb))))
	ent_write(ent);
    if ((result)!=(success))
      {
	set_skey_len(respkt, recon_this_key(blk, match_pos(pkt), key_str, 0, 0x100));
	return result;
      }
    else if (((match_type(opkt))==(pastend))
        && (!(end_of_chain_P(blk))))
      {
	set_skey_len(respkt, recon_this_key(blk, match_pos(pkt), key_str, 0, 0x100));
	return notpres;
      }
    else {
      set_skey_len(respkt, 0);
      return success;
    }
  }
}

/*  SCAN-LOOP returns SUCCESS if it reaches the end of the range, */
/*  else an ERROR code if terminated before that point, either */
/*  by an error or by FUNC returning TERMINATED. */
/*  SCAN-LOOP returns a value>0 to signal the case of */
/*  a MODIFY that requires a block-split. That value is the */
/*  length of the new value (which must be >0 to have caused an */
/*  increase in block size). SCAN-LOOP NEVER returns NOTPRES. */
/*  Note that (MATCH-POS pkt) is always the current scan point. */


int scan_loop(blk, operation, pkt, opkt, func, long_tab, respkt, ckstr, clen, vstr, blksize)
     unsigned char *blk;
     int operation;
     int *pkt;
     int *opkt;
     int_function func;
     long *long_tab;
     int *respkt;
     unsigned char *ckstr;
     int clen;
     unsigned char *vstr;
     int blksize;
{
L_scan_loop:
  if ((match_pos(opkt))>(match_pos(pkt)))
    {
      int old_bend = blk_end(blk);
      int next_pos = next_cnvpair(blk, match_pos(pkt));
      int result = success;
      if (func)
	{
	  int vpos = next_field(blk, 1+(match_pos(pkt)));
	  int vlen = field_len(blk, vpos);
	  substring_move(blk, (vpos)+1, (vpos)+(vlen)+1, vstr, 0);
	  result = func(ckstr, clen, vstr, vlen, long_tab);
	}
      if ((result)>=(success))
	{
	  if ((operation)==(rem_scan))
	    {
	      blk_remove_key_and_val(blk, match_pos(pkt), blksize);
	      set_skey_count(respkt, (skey_count(respkt))+1);
	      rem_ct = 0x1L+(rem_ct);
	      if ((match_pos(opkt))==(next_pos))
		{
		  set_match_pos(opkt, match_pos(pkt));
		}
	      else set_match_pos(opkt, (match_pos(opkt))-((old_bend)-(blk_end(blk))));
	      next_pos = match_pos(pkt);
	    }
	  else if ((operation)==(count_scan))
	    {
	      set_skey_count(respkt, (skey_count(respkt))+1);
	      set_match_pos(pkt, next_pos);
	    }
	  else if (change_existing_value(blk, match_pos(pkt), ckstr, clen, vstr, result, blksize))
	    {
	      set_skey_count(respkt, (skey_count(respkt))+1);
	      next_pos = (next_pos)-((old_bend)-(blk_end(blk)));
	      set_match_pos(opkt, (match_pos(opkt))-((old_bend)-(blk_end(blk))));
	      set_match_pos(pkt, next_pos);
	      result = success;
	    }
	  else dprintf((diagout, "%s: hit modify special case\n", "scan_loop"));
	}
      else if ((result)==(notpres))
	set_match_pos(pkt, next_pos);
      if (((result)==(success))
	  || ((result)==(notpres)))
	{
	  if (func)
	    {
	      clen = (field_len(blk, next_pos))+(field_len(blk, 1+(next_pos)));
	      substring_move(blk, (next_pos)+2, (next_pos)+2+(field_len(blk, 1+(next_pos))), ckstr, field_len(blk, next_pos));
	    }
	  goto L_scan_loop;
	}
      else return result;
    }
  else return success;
}
